﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using xcore_tools;
using xcore_tools.Data;
using xcore_tools.Common;

namespace hcore_assembler_library
{  
    class HcorePreprocessor : IPrint
    {
        IPrint m_iprint;

        public HcorePreprocessor(IPrint iprint)
        {
            m_iprint = iprint;
        }

        // defines (will used by assembler later)

        public Dictionary<string, string> defines = new Dictionary<string, string>();

        // macros

        Dictionary<string, PreprocessorMacro> macros = new Dictionary<string, PreprocessorMacro>();

        // preprocess reg exp

        const string m_cmd = @"^(?<label>[A-Za-z][A-Za-z0-9_]+)?[\s]+(?<cmd>\.[A-Za-z][A-Za-z0-9_]*)[\s]*(?<comment>;[^\n]*)?$";
        const string m_cmd_string = @"^(?<label>[A-Za-z][A-Za-z0-9_]+)?[\s]+(?<cmd>\.[A-Za-z][A-Za-z0-9_]*)[\s]+\""(?<name>[^\""]+)\""[\s]*(?<comment>;[^\n]*)?$";
        const string m_cmd_ident = @"^(?<label>[A-Za-z][A-Za-z0-9_]+)?[\s]+(?<cmd>\.[A-Za-z][A-Za-z0-9_]*)[\s]+(?<text>[A-Za-z][A-Za-z0-9]*)[\s]*(?<comment>;[^\n]*)?$";
        const string m_cmd_ident_integer = @"^(?<label>[A-Za-z][A-Za-z0-9_]+)?[\s]+(?<cmd>\.[A-Za-z][A-Za-z0-9_]*)[\s]+(?<text>[A-Za-z][A-Za-z0-9]*)[\s]+(?<integer>[\-\+$%0-9][A-Fa-f0-9]*)[\s]*(?<comment>;[^\n]*)?$";
        const string m_cmd_br_data_br = @"^(?<label>[A-Za-z][A-Za-z0-9_]+)?[\s]+(?<cmd>\.[A-Za-z][A-Za-z0-9_]*)[\s]+(?<text>[A-Za-z][A-Za-z0-9]*)[\s]*\([\s]*((?<data>[^,^\s^;]+))[\s]*([,](?<dataN>[^,^;]+)[\s]*)*[\s]*\)[\s]*(?<comment>;[^\n]*)?$";
        const string m_cmd_data = @"^(?<label>[A-Za-z][A-Za-z0-9_]+)?[\s]+(?<cmd>\.[A-Za-z][A-Za-z0-9_]*)[\s]+((?<data>[^,^\s^;]+))[\s]*([,](?<dataN>[^,^;]+)[\s]*)*[\s]*(?<comment>;[^\n]*)?$";

        Regex l1Cmd = new Regex(m_cmd, RegexOptions.Compiled);
        Regex l1CmdString = new Regex(m_cmd_string, RegexOptions.Compiled);
        Regex l1CmdIdent = new Regex(m_cmd_ident, RegexOptions.Compiled);
        Regex l1CmdIdentInteger = new Regex(m_cmd_ident_integer, RegexOptions.Compiled);
        Regex l1CmdNameBrDataBr = new Regex(m_cmd_br_data_br, RegexOptions.Compiled);
        Regex l1CmdNameData = new Regex(m_cmd_data, RegexOptions.Compiled);

        // level 2

        // process
        int nesting = 0;
        int ifdefNestingCount = 0;

        public void process(Dictionary<string, string> preDefines, List<Line> lines)
        {

            nesting = 0;
            ifdefNestingCount = 0;
            defines = preDefines; // reuse defines from caller
            macros.Clear();

            // preprocess

            dprint("Preprocess (");
            internel_process(0, lines);
            dprint(") found " + defines.Count + " identifier\n");

            // cleanup

            for (int i = 0; i < lines.Count; i++)
            {
                Line line = lines[i];
                if (line.valid == false)
                {
                    lines.RemoveAt(i);
                    i--;
                }
            }
        }

        string m_incdir = "";

        private void internel_process(int start, List<Line> lines, bool markAsInvalid = false)
        {
            dprint(".");
            nesting++;

            for (int lineNumber = start; lineNumber < lines.Count; lineNumber++)
            {
                if (lines[lineNumber].valid == false)
                {
                    continue;
                }

                if (markAsInvalid)
                {
                    lines[lineNumber].valid = false;
                }

                string lineContent = lines[lineNumber].value;

                Match match;

                if ((match = l1Cmd.Match(lineContent)).Success)  // test cmd
                {
                    string label = match.Groups["label"].Value;
                    string cmd = match.Groups["cmd"].Value;

                    switch (cmd)
                    {
                        case ".endif":
                            {
                                ifdefNestingCount--;
                                lines[lineNumber].valid = false; // remove line
                                goto stopLoop;
                            }
                        default: break;
                    }
                }
                else if ((match = l1CmdIdent.Match(lineContent)).Success)   // test cmd ident
                {
                    string label = match.Groups["label"].Value;
                    string cmd = match.Groups["cmd"].Value;
                    string ident = match.Groups["text"].Value;

                    switch (cmd)
                    {
                        case ".macro":  // macro without parameters
                            {
                                lines[lineNumber].valid = false; // remove line

                                string macroName = ident;
                                string[] macroArguments = null;

                                lineNumber = recordMacro(lines, ref lineNumber, macroName, macroArguments);
                            }
                            break;
                        case ".execm": // exec macro without parameters
                            {
                                lines[lineNumber].valid = false; // remove line

                                if (macros.ContainsKey(ident))
                                {
                                    PreprocessorMacro realMacro = macros[ident];

                                    // get arguments

                                    string[] arguments = null;
                                    string[] newLines = realMacro.resolve(arguments);
                                    int incLine = 1;
                                    foreach (string s in newLines)
                                    {
                                        lines.Insert(lineNumber + incLine, new Line(s+"\r\n", "resolved macro '" + ident + "'", (uint)lineNumber));
                                        incLine++;
                                    }
                                }
                                else
                                {
                                    throw new Exception("Error: Unknown macro named '" + ident + "' at " + lines[lineNumber].printable());
                                }
                            }
                            break;
                        case ".def":    // def without value
                            {
                                if (!markAsInvalid) // dont check in invalid lines
                                {
                                    if (defines.ContainsKey(ident))
                                    {
                                        throw new Exception("Multiple definition of define named '" + ident + "' at " + lines[lineNumber].printable());
                                    }

                                    defines.Add(ident, "");
                                    lines[lineNumber].valid = false; // remove line
                                }
                            }
                            break;
                        case ".undef":
                            {
                                if (!markAsInvalid) // dont check in invalid lines
                                {
                                    defines.Remove(ident);
                                    lines[lineNumber].valid = false; // remove line
                                }
                            }
                            break;
                        case ".ifdef":
                            {
                                bool found = defines.ContainsKey(ident);
                                lines[lineNumber].valid = false; // remove line

                                if (found && markAsInvalid)
                                {
                                    found = false;
                                }

                                ifdefNestingCount++;
                                internel_process(lineNumber + 1, lines, !found);
                            }
                            break;
                        case ".ifndef":
                            {
                                bool found = !defines.ContainsKey(ident);
                                lines[lineNumber].valid = false; // remove line

                                if (found && markAsInvalid)
                                {
                                    found = false;
                                }

                                ifdefNestingCount++;
                                internel_process(lineNumber + 1, lines, !found);
                            }
                            break;
                        default: break;
                    }
                }
                else if ((match = l1CmdIdentInteger.Match(lineContent)).Success) // test cmd indent integer
                {
                    string label = match.Groups["label"].Value;
                    string cmd = match.Groups["cmd"].Value;
                    string ident = match.Groups["text"].Value;
                    string integer = match.Groups["integer"].Value;

                    switch (cmd)
                    {
                        case ".def":    // def with value
                            {
                                if (!markAsInvalid) // dont check in invalid lines
                                {
                                    if (defines.ContainsKey(ident))
                                    {
                                        throw new Exception("Multiple definition of define named '" + ident + "' at " + lines[lineNumber].printable());
                                    }

                                    defines.Add(ident, integer);
                                    lines[lineNumber].valid = false; // remove line
                                }
                            }
                            break;
                        default: break;
                    }
                }
                else if ((match = l1CmdString.Match(lineContent)).Success)   // test cmd string
                {
                    string label = match.Groups["label"].Value;
                    string cmd = match.Groups["cmd"].Value;
                    string name = match.Groups["name"].Value;

                    switch (cmd)
                    {
                        case ".incdir":
                            {
                                m_incdir = name;
                                lines[lineNumber].valid = false; // remove line
                            }
                            break;
                        case ".incbin":
                            {
                                string fileName = m_incdir + @"\" + name;

                                // read data

                                using (FileStream fs = File.Open(fileName, FileMode.Open, FileAccess.Read, FileShare.None))
                                {
                                    long lengthByte = ((fs.Length & 1) == 1) ? fs.Length + 1 : fs.Length;   // make even
                                    long lengthWord = lengthByte / 2;

                                    UInt16[] dataRawWords = new UInt16[lengthWord];

                                    for (int j = 0; j < lengthWord; j++)
                                    {
                                        int byteA = fs.ReadByte();
                                        int byteB = fs.ReadByte();

                                        dataRawWords[j] = (UInt16)((byteA << 8) | byteB);
                                    }



                                    Line lt = new Line("", 0, dataRawWords); // special usage for better speed  
                                    lines[lineNumber] = lt; // replace 
                                    fs.Close();
                                }
                            }
                            break;

                        case ".include":
                            {
                                string incFileName = m_incdir + "\\" + name;
                                string[] rlines = File.ReadAllLines(incFileName);
                                int incLine = 1;
                                foreach (string s in rlines)
                                {
                                    lines.Insert(lineNumber + incLine, new Line(s, incFileName, (uint)incLine));
                                    incLine++;
                                }
                                lines[lineNumber].valid = false; // remove line
                            }
                            break;
                        // special for ucore
                        case ".string":
                            {
                                lines[lineNumber].valid = false; // remove line
                                string arg = name;
                                for (int j = 0; j < arg.Length; j++)
                                {
                                    string singeleChar = string.Format("{0}\tword\t${1:X4}\t;generated by string\r\n", j == 0 ? label : "", (UInt16)arg.ElementAt(j));
                                    Line lt = new Line(singeleChar, "", 0);
                                    lines.Insert(lineNumber + 1 + j, lt);
                                }
                            }
                            break;
                        case ".stringz":
                            {
                                lines[lineNumber].valid = false; // remove line
                                string arg = name + "\0";
                                int j = 0;

                                for (; j < arg.Length; j++)
                                {
                                    string singeleChar = string.Format("{0}\tword\t${1:X4}\t;generated by stringz\r\n", j == 0 ? label : "", (UInt16)arg.ElementAt(j));
                                    Line lt = new Line(singeleChar, "", 0);
                                    lines.Insert(lineNumber + 1 + j, lt);
                                }
                            }
                            break;
                        case ".stringb":
                            {
                                lines[lineNumber].valid = false; // remove line
                                string arg = name;
                                if ((arg.Length & 1) == 1) // make even if odd
                                {
                                    arg += (char)0;
                                }

                                for (int j = 0; j < arg.Length; j += 2)
                                {
                                    string singeleChar = string.Format("{0}\tword\t${1:X4}\t;generated by stringb\r\n", j == 0 ? label : "", ((((UInt16)arg.ElementAt(j)) & 0xff) << 8) | (UInt16)(arg.ElementAt(j + 1) & 0xff));
                                    Line lt = new Line(singeleChar, "", 0);
                                    lines.Insert(lineNumber + 1 + (j / 2), lt);
                                }
                            }
                            break;
                        case ".stringbz":
                            {
                                lines[lineNumber].valid = false; // remove line
                                string arg = name + "\0";
                                if ((arg.Length & 1) == 1) // make even if odd
                                {
                                    arg += (char)0;
                                }

                                for (int j = 0; j < arg.Length; j += 2)
                                {
                                    string singeleChar = string.Format("{0}\tword\t${1:X4}\t;generated by stringbz\r\n", j == 0 ? label : "", ((((UInt16)arg.ElementAt(j)) & 0xff) << 8) | (UInt16)(arg.ElementAt(j + 1) & 0xff));
                                    Line lt = new Line(singeleChar, "", 0);
                                    lines.Insert(lineNumber + 1 + (j / 2), lt);
                                }
                            }
                            break;
                        default:
                            //ignore
                            break;
                    }
                }
                else if ((match = l1CmdNameBrDataBr.Match(lineContent)).Success) // test cmd (data,...)
                {
                    string label = match.Groups["label"].Value;
                    string cmd = match.Groups["cmd"].Value;
                    string ident = match.Groups["text"].Value;
                    string data1 = match.Groups["data"].Value;
                    string comment = match.Groups["comment"].Value;
                    CaptureCollection dataN = match.Groups["dataN"].Captures;

                    string[] datas = new string[dataN.Count + 1];
                    datas[0] = data1;
                    int i = 1;
                    foreach (Capture ca in dataN)
                    {
                        datas[i] = ca.Value;
                        i++;
                    }

                    switch (cmd)
                    {
                        case ".macro":  // macro with parameters
                            {
                                lines[lineNumber].valid = false; // remove line

                                string macroName = ident;
                                lineNumber = recordMacro(lines, ref lineNumber, macroName, datas);
                            }
                            break;
                        case ".execm": // exec macro with parameters
                            {
                                lines[lineNumber].valid = false; // remove line

                                if (macros.ContainsKey(ident))
                                {
                                    PreprocessorMacro realMacro = macros[ident];

                                    // get arguments

                                    string[] arguments = datas;
                                    string[] newLines = realMacro.resolve(arguments);
                                    int incLine = 1;
                                    foreach (string s in newLines)
                                    {
                                        lines.Insert(lineNumber + incLine, new Line(s + "\r\n", "resolved macro '" + ident + "'", (uint)lineNumber));
                                        incLine++;
                                    }
                                }
                                else
                                {
                                    throw new Exception("Error: Unknown macro named '" + ident + "' at " + lines[lineNumber].printable());
                                }
                            }
                            break;
                        default: break;
                    }
                }
                else if ((match = l1CmdNameData.Match(lineContent)).Success)   // test cmd data
                {
                    string label = match.Groups["label"].Value;
                    string cmd = match.Groups["cmd"].Value;
                    string data1 = match.Groups["data"].Value;
                    string comment = match.Groups["comment"].Value;
                    CaptureCollection dataN = match.Groups["dataN"].Captures;

                    List<string> datas = new List<string>();
                    datas.Add(data1);
                    foreach (Capture ca in dataN)
                    {
                        datas.Add(ca.Value);
                    }

                    switch (cmd)
                    {
                        case ".data":
                            {
                                lines[lineNumber].valid = false; // remove line
                                for (int j = 0; j < datas.Count(); j++)
                                {
                                    string st = datas[j];
                                    string singeleWord = string.Format("{0}\tword\t{1}\t;generated by .data {2}\r\n", j == 0 ? label : "", st, comment);
                                    lines.Insert(lineNumber + 1 + j, new Line(singeleWord, "", 0));
                                }
                            }
                            break;
                        case ".datauc":
                            {
                                lines[lineNumber].valid = false; // remove line
                                for (int j = 0; j < datas.Count(); j++)
                                {
                                    UInt32 st = Convert.ToUInt32(datas[j]);
                                    st &= 0xffff;
                                    string singeleWord = string.Format("{0}\tword\t{1}\t;generated by .data {2}\r\n", j == 0 ? label : "", st, comment);
                                    lines.Insert(lineNumber + 1 + j, new Line(singeleWord, "", 0));
                                }
                            }
                            break;
                        case ".space":
                            {
                                lines[lineNumber].valid = false; // remove line                                
                                uint space = (uint)Tools.extractInteger(data1);

                                for (int j = 0; j < space; j++)
                                {
                                    string singeleWord = string.Format("\tword\t${0:X4}\t;generated by space\r\n", 0);
                                    lines.Insert(lineNumber + 1 + j, new Line(singeleWord, "", 0));
                                }
                            }
                            break;
                        default: break;
                    }
                }
            }
        stopLoop:
            nesting--;

            if (nesting == 0 && ifdefNestingCount != 0)
            {
                throw new Exception("Error in preprocessor .ifdef/.infdef has no .endif or otherwise (count = " + ifdefNestingCount + ")");
            }
        }

        private int recordMacro(List<Line> lines, ref int lineNumber, string macroName, string[] macroArguments)
        {
            String macroContent = "";

            bool foundEndMacro = false;
            for (lineNumber++; lineNumber < lines.Count; lineNumber++)
            {
                string macroInline = lines[lineNumber].value;
                Match internalMatch = l1Cmd.Match(macroInline);
                if (internalMatch.Success)
                {
                    string mcmd = internalMatch.Groups["cmd"].Value;
                    if (mcmd == ".endmacro")
                    {
                        foundEndMacro = true;
                        lines[lineNumber].valid = false; // remove line
                        break;
                    }
                }
                else
                {
                    macroContent += lines[lineNumber].value;
                    macroContent += "\n";
                }

                lines[lineNumber].valid = false; // remove line
            }

            if (foundEndMacro == false)
            {
                throw new Exception("Error: Can´t find .endmacro for macro definition '" + macroName + "'");
            }

            macros.Add(macroName, new PreprocessorMacro(macroName, macroArguments, macroContent));
            return lineNumber;
        }



        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
    }


}
