﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ucore_emulator.core.components.stages
{
    public interface IStageSignals<SELF>
    {
        void reset();
        string trace();
        void copy(ref SELF dout);        
    }

    public interface IHStageSignals<SELF>
    {
        void reset();
        string trace();
        void copy(ref SELF dout,bool stall);
    }

    public class m0_out_m1_in : IStageSignals<m0_out_m1_in>
    {
        public UInt16 address;

        public m0_out_m1_in()
        {
            reset();
        }

        public void copy(ref m0_out_m1_in dout)
        {
            dout.address = address;
        }
        
        public string trace()
        {
            return string.Format("-- M0:\nAddress: {0:X4}\n", address);
        }

        public void reset()
        {
            address = 0;
        }
    }

    public class s0_out_s1_in : IStageSignals<s0_out_s1_in>
    {
        public UInt16 address;

        public s0_out_s1_in()
        {
            reset();
        }

        public UInt16 debug_pc
        {
            get { return address; }
        }

        public void copy(ref s0_out_s1_in dout)
        {
            dout.address = address;
        }

        public string trace()
        {
            return string.Format("-- S0:\nAddress: {0:X4}\n", address);
        }

        public void reset()
        {
            address = 0;
        }
    }

    public class s1_out_s2_in : IStageSignals<s1_out_s2_in>
    {
        public UInt16 data_in;
        public UInt16 debug_pc;

        public s1_out_s2_in()
        {
            reset();
        }

        public void copy(ref s1_out_s2_in dout)
        {
            dout.data_in = data_in;
            dout.debug_pc = debug_pc;
        }

        public string trace()
        {
            return string.Format("-- S1:\nData: {0:X4}\n", data_in);
        }

        public void reset()
        {
            data_in = 0xffff;
            debug_pc = 0;
        }
    }

    public class s2_out_s3_in : IStageSignals<s2_out_s3_in>
    {
        public opcode_t decode_opcode;
        public int decode_ra;
        public int decode_rb;
        public int decode_rc;
        public bool decode_rc_we;
        public bool decode_imm_a;
        public UInt16 decode_imm;
        public UInt16 debug_pc;

        public s2_out_s3_in()
        {
            reset();
        }

        public void copy(ref s2_out_s3_in dout)
        {
            dout.decode_opcode = decode_opcode;
            dout.decode_ra = decode_ra;
            dout.decode_rb = decode_rb;
            dout.decode_rc = decode_rc;
            dout.decode_rc_we = decode_rc_we;
            dout.decode_imm_a = decode_imm_a;
            dout.decode_imm = decode_imm;
            dout.debug_pc = debug_pc;
        }

        public string trace()
        {
            return string.Format("-- S2:\nOpcode: {0}\nra: {1}\nrb: {2}\nrc: {3}\nrc_we: {4}\nimm_a: {5}\nimm: {6:X4}\n", decode_opcode.ToString(), decode_ra,decode_rb, decode_rc, decode_rc_we, decode_imm_a, decode_imm);
        }

        public void reset()
        {
            decode_opcode = opcode_t.ope_nop;
            decode_rc_we = false;
            debug_pc = 0;
        }
    }

    public class s3_out_s4_in : IStageSignals<s3_out_s4_in>
    {
        public int reg_rc;
        public opcode_t opcode;
        public UInt16 reg_data_a;
        public UInt16 reg_data_b;
        public UInt16 mem_q;
        public bool reg_rc_we;       
        public bool irq;
        public UInt16 debug_pc;
        public UInt16 simm;

        public s3_out_s4_in()
        {
            reset();
        }

        public void copy(ref s3_out_s4_in dout)
        {
            dout.opcode = opcode;
            dout.reg_data_a = reg_data_a;
            dout.reg_data_b = reg_data_b;
            dout.mem_q = mem_q;
            dout.reg_rc_we = reg_rc_we;
            dout.reg_rc = reg_rc;
            dout.simm = simm;
            dout.irq = irq;
            dout.debug_pc = debug_pc;
        }

        public string trace()
        {
            return string.Format("-- S3:\nOpcode: {0}\nrda: {1:X4}\nrdb: {2:X4}\nrdc: {3:X4}\nrdc_we: {4:X4}\nmem_q: {5:X4}\n", opcode.ToString(), reg_data_a, reg_data_b, reg_rc, reg_rc_we, mem_q);
        }

        public void reset()
        {
            opcode = opcode_t.ope_nop;
            reg_rc_we = false;
            irq = false;
            debug_pc = 0;
        }
    }

    public class s4_out_s5_in : IStageSignals<s4_out_s5_in>
    {
        public int alu_rc;
        public opcode_t opcode; // only for debug
        public UInt16 pc;
        public UInt16 alu_data_c;      
        public UInt16 mem_data;
        public UInt16 mem_address;
        public bool mem_we;
        public bool alu_rc_we;
        public UInt16 debug_pc;

        public s4_out_s5_in()
        {
            reset();
        }

        public void copy(ref s4_out_s5_in dout)
        {
            dout.pc = pc;
            dout.alu_data_c = alu_data_c;
            dout.alu_rc = alu_rc;
            dout.mem_data = mem_data;
            dout.mem_address = mem_address;
            dout.mem_we = mem_we;
            dout.alu_rc_we = alu_rc_we;
            dout.opcode = opcode;
            dout.debug_pc = debug_pc;
        }
       
        public string trace()
        {
            return string.Format("-- S4:\npc: {0:X4}\nrd: {1}\nrdc: {2:X4}\nrdc_we: {3:X4}\nmem_data: {4:X4}\nmem_adr: {5:X4}\nmem_we: {6}\n",
                pc, alu_rc, alu_data_c, alu_rc_we, mem_data, mem_address, mem_we);
        }

        public void reset()
        {
            debug_pc = pc = 0;
            alu_rc_we = false;
            mem_we = false;
        }
    }

    public class s5_out : IStageSignals<s5_out>
    {
        public UInt16 debug_pc;

        public s5_out()
        {
            reset();
        }

        public void copy(ref s5_out dout)
        {
            dout.debug_pc = debug_pc;
        }

        public string trace()
        {
            return "";
        }

        public void reset()
        {
            debug_pc = 0;
        }
    }
}
